#include "sampler.h"
#include <math.h>
#include <stdint.h>

#ifdef DBENCH
#include "test/cpucycles.h"
extern const uint64_t timing_overhead;
extern uint64_t *tred, *tadd, *tmul, *tround, *tsample, *tpack;
#define DBENCH_START() uint64_t time = cpucycles()
#define DBENCH_STOP(t) t += cpucycles() - time - timing_overhead
#else
#define DBENCH_START()
#define DBENCH_STOP(t)
#endif

/*************************************************
 * Name:        rej_uniform
 *
 * Description: Sample uniformly random coefficients in [0, Q-1] by
 *              performing rejection sampling on array of random bytes.
 *
 * Arguments:   - int32_t *a: pointer to output array (allocated)
 *              - unsigned int len: number of coefficients to be sampled
 *              - const uint8_t *buf: array of random bytes
 *              - unsigned int buflen: length of array of random bytes
 *
 * Returns number of sampled coefficients. Can be smaller than len if not enough
 * random bytes were given.
 **************************************************/
unsigned int rej_uniform(int32_t *a, unsigned int len, const uint8_t *buf,
                         unsigned int buflen) {
    unsigned int ctr, pos;
    uint32_t t;
    DBENCH_START();

    ctr = pos = 0;
    while (ctr < len && pos + 2 <= buflen) {
        t = buf[pos++];
        t |= (uint32_t)buf[pos++] << 8;
        t &= 0x3FFF;

        if (t < Q)
            a[ctr++] = t;
    }

    DBENCH_STOP(*tsample);
    return ctr;
}

/*************************************************
 * Name:        rej_eta
 *
 * Description: Sample uniformly random coefficients in [-ETA, ETA] by
 *              performing rejection sampling on array of random bytes.
 *
 * Arguments:   - int32_t *a: pointer to output array (allocated)
 *              - unsigned int len: number of coefficients to be sampled
 *              - const uint8_t *buf: array of random bytes
 *              - unsigned int buflen: length of array of random bytes
 *
 * Returns number of sampled coefficients. Can be smaller than len if not enough
 * random bytes were given.
 **************************************************/
unsigned int rej_eta(int32_t *a, unsigned int len, const uint8_t *buf,
                     unsigned int buflen) {
    unsigned int ctr, pos;
    uint32_t t0, t1;
    DBENCH_START();

    ctr = pos = 0;
    while (ctr < len && pos < buflen) {
        t0 = buf[pos] & 0x0F;
        t1 = buf[pos++] >> 4;

#if ETA == 1
        if (t0 < 15) {
            t0 = t0 - (342 * t0 >> 10) * 3;
            a[ctr++] = 1 - t0;
        }
        if (t1 < 15 && ctr < len) {
            t1 = t1 - (342 * t1 >> 10) * 3;
            a[ctr++] = 1 - t1;
        }
#elif ETA == 2
        if (t0 < 15) {
            t0 = t0 - (205 * t0 >> 10) * 5;
            a[ctr++] = 2 - t0;
        }
        if (t1 < 15 && ctr < len) {
            t1 = t1 - (205 * t1 >> 10) * 5;
            a[ctr++] = 2 - t1;
        }
#endif
    }

    DBENCH_STOP(*tsample);
    return ctr;
}
/*************************************************
 * Name:        sampler_uniform32
 *
 * Description: Sample uniformly in [0, 2^32 - 1] by using buffer
 *
 * Arguments:   - uint8_t *buf: randombytes buffer, assumed to have length more
 *                than 4
 *
 * Returns sampled number in [0, 2 ^ 32 - 1]
 **************************************************/
uint32_t sampler_uniform32(const uint8_t *buf) { return *((uint32_t *)buf); }

/*************************************************
 * Name:        sampler_uniformdbl
 *
 * Description: Sample uniformly in [0, 1) by using buffer
 *
 * Arguments:   - uint8_t *buf: randombytes buffer, assumed to have length more
 *                than 4
 *
 * Returns sampled number in [0, 1)
 **************************************************/
double sampler_uniformdbl(const uint8_t *buf) {
    return sampler_uniform32(buf) * (0x1.0p-32); // [0, 2 ^ 32 - 1] * (2 ^ -32)
}

/*************************************************
 * Name:        sampler_gaussian
 *
 * Description: Sample gaussian distribution of \sigma = \sqrt(1 / 2)
 *
 * Arguments:   - double *a: pointer to output array (allocated)
 *              - unsigned int len: number of coefficients to be sampled
 *              - const uint8_t *buf: array of random bytes
 *              - unsigned int buflen: length of array of random bytes
 *
 * Returns number of sampled coefficients. Can be smaller than len if not enough
 * random bytes were given.
 **************************************************/
// clang-format off
double xtable[256] = {1.8210885857813604e-09, 1.7015975366164973e-09, 1.6061953727907649e-09, 1.5461094369365955e-09, 1.5015597697569997e-09, 1.4658501788591972e-09, 1.4358787480751722e-09, 1.4099468438744516e-09, 1.3870202376888337e-09, 1.3664210528176688e-09, 1.3476801354500526e-09, 1.3304588994352263e-09, 1.3145047225181215e-09, 1.2996239466601647e-09, 1.2856647396832644e-09, 1.2725057820994462e-09, 1.260048549679603e-09, 1.2482118994396858e-09, 1.2369281787316986e-09, 1.2261403694895949e-09, 1.215799953084823e-09, 1.20586528755179e-09, 1.1963003560350483e-09, 1.1870737887568696e-09, 1.1781580895929853e-09, 1.169529017818484e-09, 1.1611650890070235e-09, 1.1530471684739569e-09, 1.1451581373506718e-09, 1.137482616212389e-09, 1.1300067347181716e-09, 1.1227179383398592e-09, 1.115604825216312e-09, 1.1086570076514437e-09, 1.10186499390625e-09, 1.0952200868069871e-09, 1.0887142963691285e-09, 1.0823402641672329e-09, 1.0760911975993646e-09, 1.0699608125271575e-09, 1.0639432830384433e-09, 1.0580331972932362e-09, 1.0522255185869467e-09, 1.0465155509055546e-09, 1.0408989083626875e-09, 1.0353714880032982e-09, 1.0299294455368825e-09, 1.0245691736281276e-09, 1.01928728242701e-09, 1.0140805820656445e-09, 1.0089460668872618e-09, 1.0038809012047657e-09, 9.988824064135036e-10, 9.939480493059463e-10, 9.890754314556364e-10, 9.842622795545747e-10, 9.795064366026213e-10, 9.748058538598905e-10, 9.701585834838008e-10, 9.65562771781694e-10, 9.610166530179478e-10, 9.565185437214806e-10, 9.52066837445615e-10, 9.476599999375733e-10, 9.432965646795254e-10, 9.38975128767176e-10, 9.34694349095472e-10, 9.304529388241564e-10, 9.262496640986944e-10, 9.220833410045493e-10, 9.179528327349853e-10, 9.138570469545031e-10, 9.097949333417451e-10, 9.057654812972429e-10, 9.017677178027552e-10, 8.978007054201637e-10, 8.938635404189997e-10, 8.899553510226561e-10, 8.860752957642218e-10, 8.822225619436732e-10, 8.783963641788733e-10, 8.745959430434725e-10, 8.708205637853907e-10, 8.670695151200867e-10, 8.633421080932965e-10, 8.596376750083577e-10, 8.559555684136261e-10, 8.522951601458517e-10, 8.486558404256992e-10, 8.450370170019023e-10, 8.414381143407997e-10, 8.378585728582558e-10, 8.342978481911884e-10, 8.307554105061304e-10, 8.272307438424433e-10, 8.237233454879696e-10, 8.202327253850694e-10, 8.167584055651317e-10, 8.132999196097828e-10, 8.098568121371384e-10, 8.064286383115519e-10, 8.030149633754257e-10, 7.996153622017354e-10, 7.962294188660116e-10, 7.92856726236606e-10, 7.894968855821353e-10, 7.861495061950741e-10, 7.828142050305264e-10, 7.794906063592648e-10, 7.761783414341818e-10, 7.728770481693481e-10, 7.695863708309163e-10, 7.663059597391557e-10, 7.630354709809407e-10, 7.597745661320512e-10, 7.565229119886811e-10, 7.532801803075785e-10, 7.500460475542727e-10, 7.468201946588678e-10, 7.436023067789101e-10, 7.40392073068857e-10, 7.371891864556959e-10, 7.339933434202835e-10, 7.308042437839902e-10, 7.276215905002531e-10, 7.244450894506542e-10, 7.212744492451541e-10, 7.18109381026124e-10, 7.149495982758305e-10, 7.117948166270336e-10, 7.08644753676373e-10, 7.054991288002187e-10, 7.02357662972674e-10, 6.992200785854202e-10, 6.960860992690989e-10, 6.929554497159312e-10, 6.898278555032743e-10, 6.867030429178181e-10, 6.835807387801264e-10, 6.804606702692232e-10, 6.773425647469291e-10, 6.742261495816416e-10, 6.711111519712607e-10, 6.679972987649466e-10, 6.64884316283398e-10, 6.61771930137328e-10, 6.586598650438117e-10, 6.555478446401657e-10, 6.524355912950134e-10, 6.493228259161778e-10, 6.462092677550285e-10, 6.430946342068989e-10, 6.399786406071706e-10, 6.368610000226077e-10, 6.337414230375012e-10, 6.306196175341663e-10, 6.274952884673087e-10, 6.243681376317541e-10, 6.212378634230054e-10, 6.181041605900634e-10, 6.149667199799128e-10, 6.118252282730405e-10, 6.086793677093147e-10, 6.055288158035085e-10, 6.0237324504971e-10, 5.992123226138041e-10, 5.960457100131639e-10, 5.928730627826219e-10, 5.896940301257329e-10, 5.86508254550265e-10, 5.833153714867768e-10, 5.801150088890589e-10, 5.769067868151184e-10, 5.736903169872879e-10, 5.704652023299282e-10, 5.672310364830695e-10, 5.639874032902069e-10, 5.607338762583132e-10, 5.574700179879754e-10, 5.541953795713809e-10, 5.509094999556842e-10, 5.476119052690704e-10, 5.443021081065921e-10, 5.409796067725929e-10, 5.376438844762414e-10, 5.342944084763763e-10, 5.309306291715024e-10, 5.275519791303834e-10, 5.241578720582347e-10, 5.207477016930221e-10, 5.173208406258271e-10, 5.138766390386232e-10, 5.104144233521184e-10, 5.069334947755476e-10, 5.034331277494347e-10, 4.999125682713643e-10, 4.963710320937097e-10, 4.928077027810165e-10, 4.892217296133433e-10, 4.856122253202606e-10, 4.819782636284098e-10, 4.783188766034544e-10, 4.746330517649122e-10, 4.709197289496652e-10, 4.671777968968616e-10, 4.63406089523385e-10, 4.5960338185497797e-10, 4.5576838557339613e-10, 4.5189974413450207e-10, 4.4799602740585955e-10, 4.4405572576499017e-10, 4.400772435908057e-10, 4.3605889207058487e-10, 4.319988812329261e-10, 4.27895311103013e-10, 4.237461618598184e-10, 4.195492828549932e-10, 4.153023803294378e-10, 4.1100300363505663e-10, 4.0664852973486484e-10, 4.022361457130462e-10, 3.9776282897599375e-10, 3.93225324763524e-10, 3.8862012051343247e-10, 3.8394341652854145e-10, 3.7919109227844455e-10, 3.743586675217258e-10, 3.694412572498559e-10, 3.644335192196243e-10, 3.593295925411265e-10, 3.5412302540154753e-10, 3.4880668950165594e-10, 3.433726781207601e-10, 3.3781218384858403e-10, 3.3211535084572626e-10, 3.2627109489721223e-10, 3.202668823282202e-10, 3.140884557912585e-10, 3.0771949060467285e-10, 3.011411590944039e-10, 2.9433157126701733e-10, 2.872650465032816e-10, 2.799111501108847e-10, 2.722333958867869e-10, 2.641874631186133e-10, 2.5571868863290726e-10, 2.4675844268014895e-10, 2.372187234471083e-10, 2.2698378476279043e-10, 2.158965621101162e-10, 2.0373538239294718e-10, 1.9017101014591838e-10, 1.746794827647475e-10, 1.5633996539490318e-10, 1.3326042880867258e-10, 1.0022981836689348e-10};
double ytable[256] = {0.0012602859304985956, 0.002609072746102159, 0.004037972593363024, 0.005522403299250989, 0.007050875471373216, 0.00861658276939872, 0.010214971439701456, 0.01184275785790787, 0.013497450601739859, 0.015177088307935302, 0.016880083152543142, 0.018605121275724616, 0.020351096230044486, 0.02211706270730883, 0.023902203305795844, 0.025705804008548855, 0.027527235669603037, 0.029365939758133265, 0.031221417191920196, 0.03309321945857847, 0.03498094146171603, 0.03688421568856723, 0.03880270740452606, 0.04073611065594087, 0.04268414491647437, 0.04464655225129438, 0.0466230949019303, 0.04861355321586845, 0.05061772386094769, 0.052635418276792106, 0.05466646132488884, 0.05671069010620282, 0.058767952920933675, 0.060838108349539774, 0.06292102443775803, 0.06501657797124276, 0.0671246538277884, 0.06924514439700667, 0.07137794905889028, 0.07352297371398117, 0.07568013035892697, 0.07784933670209594, 0.08003051581466294, 0.08222359581320275, 0.08442850957035326, 0.08664519445055785, 0.08887359206827568, 0.09111364806637352, 0.09336531191269075, 0.09562853671300871, 0.09790327903886217, 0.1001894987688097, 0.10248715894193497, 0.10479622562248678, 0.10711666777468351, 0.1094484571468115, 0.11179156816383787, 0.11414597782783821, 0.11651166562561066, 0.11888861344290982, 0.12127680548479006, 0.12367622820159639, 0.1260868702201857, 0.12850872227999935, 0.13094177717364414, 0.13338602969166893, 0.13584147657125353, 0.1383081164485505, 0.14078594981444448, 0.1432749789735132, 0.14577520800599383, 0.14828664273257433, 0.15080929068184548, 0.15334316106026263, 0.155888264724479, 0.1584446141559241, 0.16101222343751087, 0.1635911082323655, 0.16618128576448185, 0.1687827748012113, 0.17139559563750573, 0.17401977008183853, 0.17665532144373475, 0.17930227452284742, 0.18196065559952232, 0.18463049242679902, 0.18731181422380003, 0.19000465167046474, 0.1927090369035889, 0.19542500351413403, 0.19815258654577486, 0.2008918224946563, 0.2036427493103346, 0.20640540639788046, 0.20917983462112474, 0.2119660763070299, 0.2147641752511733, 0.21757417672433085, 0.22039612748015167, 0.22323007576391715, 0.2260760713223799, 0.22893416541467992, 0.23180441082433828, 0.23468686187232957, 0.23758157443123762, 0.24048860594050006, 0.24340801542274978, 0.2463398635012633, 0.2492842124185279, 0.25224112605594157, 0.25521066995466135, 0.25819291133761857, 0.2611879191327205, 0.2641957639972604, 0.26721651834356075, 0.27025025636587485, 0.2732970540685765, 0.2763569892956677, 0.2794301417616374, 0.2825165930837071, 0.2856164268155012, 0.28872972848218237, 0.29185658561709465, 0.2949970877999613, 0.298151326696685, 0.30131939610080255, 0.30450139197664944, 0.3076974125042915, 0.31090755812628595, 0.3141319315963367, 0.31737063802991305, 0.32062378495690486, 0.3238914823763906, 0.32717384281360085, 0.3304709813791629, 0.33378301583071773, 0.3371100666370054, 0.34045225704452114, 0.34380971314685005, 0.3471825639567929, 0.35057094148140533, 0.353974980800076, 0.35739482014577967, 0.36083060098964714, 0.3642824681290031, 0.36775056977903164, 0.37123505766823856, 0.3747360871378902, 0.37825381724561824, 0.3817884108733927, 0.38534003484007634, 0.3889088600187878, 0.3924950614593147, 0.3960988185158315, 0.39972031498019633, 0.4033597392211136, 0.4070172843294725, 0.4106931482701873, 0.41438753404089024, 0.4181006498378473, 0.421832709229495, 0.4255839313380211, 0.4293545410294406, 0.4331447691126514, 0.4369548525479846, 0.44078503466580304, 0.4446355653957384, 0.448506701507202, 0.4523987068618475, 0.4563118526787154, 0.46024641781284176, 0.4642026890481732, 0.46818096140569243, 0.472181538467729, 0.4762047327195047, 0.4802508659090456, 0.48432026942668205, 0.48841328470545675, 0.49253026364386726, 0.49667156905248844, 0.5008375751261475, 0.5050286679434669, 0.5092452459957466, 0.5134877207473255, 0.5177565172297549, 0.5220520746723204, 0.5263748471716829, 0.5307253044036605, 0.5351039323804561, 0.5395112342569505, 0.5439477311900246, 0.5484139632552641, 0.5529104904258306, 0.5574378936187643, 0.5619967758145227, 0.5665877632561627, 0.5712115067352515, 0.5758686829723519, 0.5805599961007891, 0.5852861792633696, 0.5900479963328241, 0.5948462437679856, 0.5996817526191235, 0.604555390697466, 0.6094680649257717, 0.614420723888912, 0.6194143606058324, 0.6244500155470246, 0.6295287799248348, 0.6346517992876217, 0.6398202774530547, 0.6450354808208204, 0.6502987431108148, 0.6556114705796954, 0.6609751477766612, 0.6663913439087481, 0.67186171989708, 0.6773880362187713, 0.6829721616449925, 0.6886160830046695, 0.6943219161261144, 0.7000919181365093, 0.7059285013327519, 0.711834248878246, 0.7178119326307195, 0.7238645334686277, 0.7299952645614737, 0.7362075981268601, 0.7425052963401485, 0.7488924472191543, 0.7553735065070935, 0.7619533468367926, 0.7686373157984835, 0.7754313049811844, 0.7823418326547996, 0.7893761435660217, 0.7965423304229561, 0.8038494831709614, 0.8113078743126533, 0.8189291916036994, 0.8267268339462184, 0.8347162929868803, 0.842915653112201, 0.8513462584586746, 0.8600336211963281, 0.8690086880368535, 0.8783096558089137, 0.8879846607558296, 0.8980959218983395, 0.9087264400521268, 0.9199915050393427, 0.9320600759592258, 0.9451989534422945, 0.9598790918001009, 0.9771017012676645, 0.9999999999999898};
int32_t ktable[256] = {2006576130, 2027082330, 2067148736, 2085605958, 2096412913, 2103575234, 2108700193, 2112564239, 2115590521, 2118030199, 2120042179, 2121732132, 2123173182, 2124417616, 2125503854, 2126460795, 2127310606, 2128070594, 2128754474, 2129373263, 2129935916, 2130449794, 2130921000, 2131354644, 2131755037, 2132125842, 2132470192, 2132790784, 2133089954, 2133369733, 2133631898, 2133878010, 2134109446, 2134327426, 2134533033, 2134727237, 2134910901, 2135084804, 2135249647, 2135406060, 2135554617, 2135695838, 2135830194, 2135958117, 2136080000, 2136196205, 2136307061, 2136412873, 2136513921, 2136610462, 2136702735, 2136790960, 2136875342, 2136956069, 2137033318, 2137107253, 2137178025, 2137245777, 2137310641, 2137372741, 2137432193, 2137489105, 2137543578, 2137595708, 2137645583, 2137693289, 2137738902, 2137782498, 2137824146, 2137863912, 2137901857, 2137938039, 2137972514, 2138005332, 2138036544, 2138066194, 2138094326, 2138120982, 2138146200, 2138170016, 2138192465, 2138213579, 2138233390, 2138251925, 2138269214, 2138285280, 2138300150, 2138313844, 2138326386, 2138337795, 2138348090, 2138357290, 2138365410, 2138372466, 2138378474, 2138383446, 2138387395, 2138390333, 2138392271, 2138393218, 2138393184, 2138392177, 2138390204, 2138387272, 2138383388, 2138378556, 2138372780, 2138366065, 2138358414, 2138349828, 2138340311, 2138329862, 2138318483, 2138306172, 2138292930, 2138278754, 2138263643, 2138247594, 2138230603, 2138212665, 2138193778, 2138173934, 2138153129, 2138131355, 2138108605, 2138084871, 2138060145, 2138034417, 2138007677, 2137979914, 2137951118, 2137921275, 2137890372, 2137858397, 2137825334, 2137791168, 2137755883, 2137719461, 2137681886, 2137643138, 2137603197, 2137562042, 2137519652, 2137476005, 2137431075, 2137384839, 2137337269, 2137288339, 2137238020, 2137186282, 2137133093, 2137078422, 2137022233, 2136964492, 2136905160, 2136844199, 2136781568, 2136717224, 2136651122, 2136583216, 2136513458, 2136441796, 2136368176, 2136292543, 2136214839, 2136135002, 2136052968, 2135968670, 2135882037, 2135792996, 2135701469, 2135607375, 2135510629, 2135411140, 2135308816, 2135203557, 2135095258, 2134983811, 2134869100, 2134751004, 2134629394, 2134504137, 2134375088, 2134242097, 2134105007, 2133963647, 2133817841, 2133667400, 2133512123, 2133351799, 2133186203, 2133015095, 2132838221, 2132655311, 2132466075, 2132270206, 2132067375, 2131857231, 2131639398, 2131413474, 2131179026, 2130935590, 2130682669, 2130419726, 2130146181, 2129861410, 2129564738, 2129255433, 2128932702, 2128595683, 2128243439, 2127874945, 2127489086, 2127084638, 2126660260, 2126214479, 2125745669, 2125252037, 2124731597, 2124182144, 2123601224, 2122986095, 2122333689, 2121640557, 2120902808, 2120116041, 2119275251, 2118374730, 2117407932, 2116367320, 2115244162, 2114028296, 2112707817, 2111268686, 2109694236, 2107964523, 2106055481, 2103937799, 2101575411, 2098923437, 2095925311, 2092508731, 2088579773, 2084014180, 2078644066, 2072236971, 2064461601, 2054829228, 2042587920, 2026518616, 2004507660, 1972547407, 1922020343, 1830463446, 1615197384, 1426};
// clang-format on
unsigned int sampler_gaussian(double *a, unsigned int len, const uint8_t *buf,
                              unsigned int buflen) {
    unsigned int ctr, pos;
    int32_t u0, t;
    uint8_t i;
    double xvar, yvar;
    DBENCH_START();
    ctr = pos = 0;
    while (ctr < len && pos + 5 < buflen) {
        i = buf[pos++];                    // (0, 255)
        u0 = sampler_uniform32(buf + pos); // [-2 ^ 31, 2 ^ 31 - 1]
        pos += 4;
        t = u0 >> 31;                  // t = 0xffffffff if u0 < 0 else 0
        t = (t ^ u0) + (t & 1);        // t = abs(u0)
        if (t < ktable[i]) {           // abs(u0) < ktable[i]
            a[ctr++] = xtable[i] * u0; // x_i * u0
        } else {
            if (i == 0) {
                while (pos + 8 < buflen) {
                    xvar = -log(sampler_uniformdbl(buf + pos)) *
                           0.2736612373297582;
                    // x = -log(uniform(0, 1)) / x_1
                    yvar = -log(sampler_uniformdbl(buf + pos + 4));
                    pos += 8;
                    if (2 * yvar > xvar * xvar) {
                        a[ctr++] =
                            u0 > 0 ? xvar + xtable[1] : -xvar - xtable[1];
                        break;
                    }
                }
            } else if (pos + 4 < buflen) {
                xvar = xtable[i] * u0;
                yvar = ytable[i - 1] + sampler_uniformdbl(buf + pos) *
                                           (ytable[i] - ytable[i - 1]);
                pos += 4;
                if (yvar < exp(-xvar * xvar / 2)) {
                    a[ctr++] = xvar;
                }
            }
        }
    }

    DBENCH_STOP(*tsample);
    return ctr;
}
